--
-- Manfredas Zabarauskas
-- University of Edinburgh
-- BSc Artificial Intelligence & Computer Science '08
--
-- Informatics 1A Competition Entry
--
-- Render.hs
--

module Render (
               Bitmap,
               RasterPoint,
               Coord,
               Color,
               
               projectPoint,           -- (Coord, Coord, Coord) -> Color -> RasterPoint
               rasterLine,             -- RasterPoint -> RasterPoint -> [RasterPoint]
               rasterRectFill,         -- (Coord, Coord) -> (Coord, Coord) -> Color -> [RasterPoint]

               renderPointsToBMP,      -- [RasterPoint] -> Bitmap -> Bitmap
               superimposePointsToBMP, -- [RasterPoint] -> Bitmap -> Bitmap

               expandBitmap,           -- Bitmap -> Int -> Bitmap
                
               init_bitmap,            -- Bitmap
               height,                 -- Int
               width,                  -- Int
               depth,                  -- Int

               -- Color
               black, grey, red, green, blue, yellow, white, purple,
               dark_blue, dark_purple, dark_cyan, cyan, light_cyan, bright_cyan

              )               
       where

import Array

type Coord = Int
type Bitmap = Array (Coord, Coord) Color

type RasterPoint = ((Coord, Coord), Color)

type Color = (Int, Int, Int)

-- Initial bitmap settings
width = 800
height = 600
depth = 400
      
init_bitmap :: Bitmap
init_bitmap = array ((0, 0), (width-1, height-1)) [((x, y), black) | x <- [0..width-1], y <- [0..height-1]]

-- Camera position and angle
deg = pi / (fromIntegral 4)
sinx = sin deg
cosx = cos deg

cx = 3.0
cy = 2.0
cz = 0.0

projectPoint :: (Coord, Coord, Coord) ->  -- 3D point
                Color ->                  -- color
                RasterPoint               -- rasterized point
projectPoint (x, y, z) clr = ((if proj_x < 0 then 0 else (if proj_x > (width - 1) then width - 1 else proj_x),
                              if proj_y < 0 then 0 else (if proj_y > (height - 1) then height - 1 else proj_y)),
                              clr)
    where proj_x = round (dx / dz * (fromIntegral width) / 2) + fromIntegral trans_x
          proj_y = round (dy / dz * (fromIntegral height) / 2) + fromIntegral trans_y

          dx = ax - cx
          dy = sinx * (az - cz) + cosx * (ay - cy)
          dz = cosx * (az - cz) - sinx * (ay - cy)
            
          ax = fromIntegral (2*(x - trans_x)) / (fromIntegral width)
          ay = fromIntegral (2*(y - trans_y)) / (fromIntegral height)
          az = (fromIntegral z) / (fromIntegral depth)

          trans_x = width `div` 2
          trans_y = height `div` 2


renderPointsToBMP :: [RasterPoint] -> -- initial points
                     Bitmap ->        -- initial bitmap
                     Bitmap           -- resulting bitmap
renderPointsToBMP =  flip (//) --filter (\((x, y), c) -> (x >= 0) && (x < width) && (y >= 0) && (y < height)) p

superimposePointsToBMP :: [RasterPoint] -> -- initial points
                          Bitmap ->        -- initial bitmap
                          Bitmap           -- resulting bitmap
superimposePointsToBMP pts bmp = renderPointsToBMP [ ((x, y), addClr (bmp!(x,y)) clr) | ((x, y), clr) <- pts ] bmp
    where addClr (r1, g1, b1) (r2, g2, b2) = (bound (r1+r2), bound (g1+g2), bound (b1+b2))
          bound n = if n > 255 then 255 else n

-- Bresenham's (Midpoint) Algorithm
rasterLine :: RasterPoint -> -- start point
              RasterPoint -> -- end point
             [RasterPoint]   -- resulting line
rasterLine start end = renderLinePoint d x0 y0
    where 
      clr = getColor start

      swapXY = (abs (getY end - getY start) > abs (getX end - getX start))    
      start'' = if swapXY then ((getY start, getX start), clr) else start
      end''   = if swapXY then ((getY end, getX end), clr)     else end
      
      swapX = (getX start'' > getX end'')
      start' = if (swapX) then end''   else start''
      end'   = if (swapX) then start'' else end''

      x0 = getX start'
      y0 = getY start'
      x1 = getX end'
      y1 = getY end'
 
      swapY = if (y0 > y1) then (-1) else 1

      dy = swapY * (y1 - y0)
      dx = x1 - x0

      incrE = 2 * dy
      incrNE = 2 * (dy - dx)     
 
      d = 2 * dy - dx

      renderLinePoint :: Int -> Int -> Int -> [RasterPoint]
      renderLinePoint d' x' y'
          | x' < x1   = ((if swapXY then (y', x') else (x', y')), clr) : (renderLinePoint (d' + if (d' <= 0) then incrE else incrNE) 
                                                                                          (x' + 1) 
                                                                                          (y' + if (d' <= 0) then 0     else swapY))
          | otherwise = [((if swapXY then (y', x') else (x', y')), clr)]


rasterRectFill :: (Coord, Coord) ->   -- left bottom corner of rectangle
                  (Coord, Coord) ->   -- right upper corner of rectangle
                  Color          ->   -- fill color
                  [RasterPoint]       -- raster rectangle
rasterRectFill (x0, y0) (x1, y1) clr = concat [ rasterLine ((x0', y), clr) ((x1', y), clr) | y <- [y0'..y1'] ]
    where 
      x0' = filter_x x0; x1' = filter_x x1
      y0' = filter_y y0; y1' = filter_y y1

      filter_x x = if (x < 0) then 0 else if (x >= width) then width - 1 else x
      filter_y y = if (y < 0) then 0 else if (y >= height) then height - 1 else y


expandBitmap :: Bitmap -> -- initial bitmap
                Int ->    -- border size
                Bitmap    -- new bitmap
expandBitmap bmp border = new_bmp // (map (\((x, y), clr) -> ((x + border, y + border), clr)) (assocs bmp))
    where ((x0, y0), (x1, y1)) = bounds bmp
          new_width = (x1 - x0) + 2*border + 1
          new_height = (y1 - y0) + 2*border + 1

          new_bmp = array ((x0, y0), (new_width - 1, new_height - 1)) [((x, y), black) | x <- [x0..new_width - 1], y <- [y0..new_height - 1]]


getX :: RasterPoint -> Coord
getX ((x, y), c) = x

getY :: RasterPoint -> Coord
getY ((x, y), c) = y

getColor :: RasterPoint -> Color
getColor ((x, y), c) = c

--------- COLOR LIST --------

black = (0, 0, 0) :: Color
grey =  (200, 200, 200) :: Color
red =   (255, 0, 0) :: Color
green = (0, 255, 0) :: Color
blue =  (0, 0, 255) :: Color
yellow = (255, 255, 0) :: Color
dark_blue = (0, 0, 160) :: Color
white = (255, 255, 255) :: Color
purple = (255, 0, 255) :: Color
dark_purple = (128, 0, 128) :: Color
dark_cyan = (0, 128, 128) :: Color
cyan = (0, 200, 200) :: Color
light_cyan = (0, 255, 255) :: Color
bright_cyan = (100, 255, 255) :: Color

