--
-- Manfredas Zabarauskas
-- University of Edinburgh
-- BSc Artificial Intelligence & Computer Science '08
--
-- Informatics 1A Competition Entry
--
-- World.hs
--

module World (
              WorldObject (..),

              generateTerrain,     -- Shape -> [Int] -> Int -> WorldObject
              generateTrees,       -- [Shape] -> [WorldObject]
              generateStars,       -- Int -> [Float] -> [RasterPoint]
              generatePodium,      -- WorldObject -> Int -> Int -> [Shape]
              placeLambda,         -- [Shape] -> [Shape]
              placeCloud,          -- Shape -> Int -> Int -> Int -> [RasterPoint]

              getMountains,        -- WorldObject -> Int -> [Shape]
              getShapeList,        -- WorldObject -> [Shape]

              lambda_top_index     -- Int
             )
       where

import List

import Graphics
import Render
import Fractal

data WorldObject = Tree [Shape]
                 | Terrain [Shape]
                 | Object [Shape]

generateStars :: Int ->        -- count
                 [Float] ->    -- random list (0..1)
                 [RasterPoint] -- generated stars
generateStars 0 _ = []
generateStars (count + 1) (r_x:r_y:rand_l) = randomStar (round (r_x * fromIntegral (width - 1))) 
                                                        (round (r_y * fromIntegral (height - 1))) ++ generateStars count rand_l
    where randomStar x y = case (x `mod` 4) of 0 -> [((x, y), clr)]
                                               1 -> [((x, y), clr), ((x+1, y), clr)]
                                               2 -> [((x, y), clr), ((x, y+1), clr)]
                                               3 -> [((x, y), clr), ((x+1, y), clr), ((x, y+1), clr), ((x+1, y+1), clr)]
          clr = white

placeCloud :: Shape ->      -- initial shape over which the cloud will be placed
              Int ->        -- distance between the shape and the cloud
              Int ->        -- cloud width
              Int ->        -- cloud height
              [RasterPoint] -- projected cloud
placeCloud (Rectangle (x',y',z') b c d clr) dist width height = (map proj_transl (generateJulia cloud width height (5,7,7)))
    where
      proj_transl ((x, z), clr') = projectPoint (x+x'-(width `div` 2), y'+dist, z+z'-(height `div` 2)) clr'


placeLambda :: [Shape] ->  -- generated podium
               [Shape]     -- translated lambda object`s shape list
placeLambda podium = getShapeList $ translateObject lambda ((\(x, y, z) -> (x - offs_x, y, z - offs_z)) b)
    where
      podium_top = [ x | (x, y) <- zip podium [1..], y `mod` 5 == 0 ]

      (Rectangle a b c d clr) = podium_top!!((length podium_top `div` 2) - 1)

      offs_x = get_X (b .-. a) `div` 4
      offs_z = get_Z (b .-. a) `div` 4


generatePodium :: WorldObject -> -- generated terrain
                  Int ->         -- height of podium
                  Int ->         -- random position
                  [Shape]        -- generated podium`s shape list
generatePodium (Terrain terr) height rand_pos = (concat . (map raiseRect)) base
    where
      base = map (\x -> terr!!((((length terr) `div` 16) * (case (rand_pos `mod` 4) of  -- randomizing location for the base of podium
                                                              0 -> 2  -- 2 4 7 10
                                                              1 -> 4
                                                              2 -> 7
                                                              3 -> 10)) - x)) [1..16]

      raiseRect (Rectangle a b c d clr) = [(Rectangle a' d' d  a  new_clr),
                                           (Rectangle d' c' c  d  new_clr),
                                           (Rectangle c' b' b  c  new_clr),
                                           (Rectangle b' a' a  b  new_clr),
                                           (Rectangle a' b' c' d' new_clr)]

          where (Rectangle a' b' c' d' clr') = (Rectangle (lift_y a) (lift_y b) (lift_y c) (lift_y d) clr)--translate (Rectangle a b c d clr) (0, height, 0)
                new_clr = light_cyan

                lift_y :: Point -> Point
                lift_y (x, y, z) = (x, base_height + height, z)
                    where
                      base_height = round (sum (map (\(Rectangle a b c d clr) -> fromIntegral (get_Y a + get_Y b + get_Y c + get_Y d) / 4.0) base) /
                                    fromIntegral (length base))


generateTrees :: [Shape] ->    -- generated mountains
                 [WorldObject] -- translated tree objects
generateTrees mountains = map (\(Rectangle a b c d clr) -> translateObject tree2 a) (map (fst) (filter (even . snd) numb_mnts)) ++
                          map (\(Rectangle a b c d clr) -> translateObject tree1 a) (map (fst) (filter (odd . snd) numb_mnts))
    where numb_mnts = zip mountains [0..]

-- Recursive Tree-Generation Algorithm
generateTree :: Shape ->    -- root line
                Int ->      -- length of branches
                Int ->      -- depth of recursive generation
                Float ->    -- angle of branches (relative to previous branch; radians)
                Color ->    -- color of leaf
                WorldObject -- generated tree
generateTree root len n deg leaf_clr = Tree (root : (generateBranches root (0, 1, 0) len n))
    where
      generateBranches :: Shape -> (Float, Float, Float) -> Int -> Int -> [Shape]
      generateBranches (Line (x1,y1,z1) (x2,y2,z2) clr) (dirX, dirY, dirZ) len' n'
          | n' > 0 = map (\x -> Line (x2,y2,z2) x clr) new_brch_pts
                     ++ concat (map (\(l, p) -> generateBranches (Line (x2,y2,z2) l clr) p new_len (n' - 1)) 
                               $ zip new_brch_pts [(dirXp, dirYp, 0), (dirXn, dirYn, 0), (0, dirYp, dirXp), (0, dirYn, dirXn)])
          | otherwise = map (\x -> Line (x2,y2,z2) x leaf_clr) new_brch_pts
          where
            new_brch_pts = [new_px, new_nx, new_pz, new_nz]

            new_px = (x2 + offs_xp, y2 + round (dirYp * fromIntegral len'), z2)
            new_nx = (x2 + offs_xn, y2 + round (dirYn * fromIntegral len'), z2)
            new_pz = (x2, y2 + round (dirYp * fromIntegral len'), z2 + offs_xp)
            new_nz = (x2, y2 + round (dirYn * fromIntegral len'), z2 + offs_xn)
      
            offs_xp = round (dirXp * fromIntegral len')
            offs_xn = round (dirXn * fromIntegral len')

            dirXp = (cos deg * dirX) + (sin deg * dirY)
            dirYp = (cos deg * dirY) - (sin deg * dirX)

            dirXn = (cos (-deg) * dirX) + (sin (-deg) * dirY)
            dirYn = (cos (-deg) * dirY) - (sin (-deg) * dirX)
              
            new_len = round (2.0/3.0 * fromIntegral len')

getMountains :: WorldObject -> -- generated terrain
                Int ->         -- number of mountains
                [Shape]        -- filtered mountains
getMountains (Terrain x) n = getMountains' x
  where 
    len = (length x) `div` n
        
    getMountains' [] = []
    getMountains' x = mountains (take len x) 1 ++ getMountains' (drop len x)

    mountains :: [Shape] -> Int -> [Shape]
    mountains shapes n = iterate (sortBy isHigher (take n shapes)) (drop n shapes)
        where
          iterate :: [Shape] -> [Shape] -> [Shape]
          iterate found [] = found
          iterate found (x:xs) | (x `isHigher` (found!!(n - 1)) == LT) = iterate (sortBy isHigher (x : init found)) xs
                               | otherwise            = iterate found xs

          isHigher (Rectangle a1 b1 c1 d1 clr1) (Rectangle a2 b2 c2 d2 clr2) 
              | (y_sum a1 b1 c1 d1) > (y_sum a2 b2 c2 d2) = LT
              | otherwise = GT
              where
                y_sum a b c d = get_Y a + get_Y b + get_Y c + get_Y d

-- Midpoint Displacement (MDP) Terrain Algorithm
--  a  _______  b
--    | 0 | 1 |
--    |---X---|
--  d |_2_|_3_| c
--
generateTerrain :: Shape ->      -- initial rectangle
                   [Int] ->      -- list of random integers
                   Int ->        -- level of detail (recursion depth)
                   WorldObject   -- generated terrain
generateTerrain rect rand_l n = Terrain (recursiveSplit [rect] n 0 rand_l 1.0)
    where
      recursiveSplit :: [Shape] -> Int -> Int -> [Int] -> Float -> [Shape]
      recursiveSplit rects 0 _ _ _ = rects
      recursiveSplit rects n i rnds delta
          | n > 0 = recursiveSplit (concat $ map (\(x, y) -> splitRectangle x y) (zip rects [0..]))
                                   (n - 1) 
                                   (i + 1) 
                                   new_rnds 
                                   new_delta
          where
            pow4_l = map (4^) [0..]
      
            heights = map (\(x, y) -> calcDisplacement x y delta) (zip rects (take (pow4_l !! i) rnds))
            
            new_rnds = drop (pow4_l !! i) rnds
            new_delta = (sqrt 2.0)/2 * delta
            
            calcDisplacement :: Shape -> Int -> Float -> Int
            calcDisplacement (Rectangle (ax, ay, az) (bx, by, bz) (cx, cy, cz) (dx, dy, dz) clr) height delta =  
                ((ay + by + cy + dy) `div` 4) + round ((fromIntegral height)*delta)

            splitRectangle :: Shape -> Int -> [Shape]
            splitRectangle (Rectangle (ax, ay, az) (bx, by, bz) (cx, cy, cz) (dx, dy, dz) clr) num = rec_l
                where
                  rec_l = [rec1, rec2, rec3, rec4]

                  rec1 = Rectangle a    a_b  x    a_d  clr
                  rec2 = Rectangle a_b  b    b_c  x    clr
                  rec3 = Rectangle a_d  x    d_c  d    clr
                  rec4 = Rectangle x    b_c  c    d_c  clr
      
                  x   = ((bx - ax) `div` 2 + ax, (heights!!num), (az - dz) `div` 2 + dz)

                  offs_y = length heights
                  offs_x = round (sqrt (fromIntegral offs_y))

                  mapping = f i
                  mapping_inv = f_i i
                  map_n_i = (mapping_inv)!!num

                  a_b = ((bx - ax) `div` 2 + ax, 
                        ((ay + get_Y x + by + (heights!!(mapping!!((map_n_i - offs_x) `mod` offs_y)))) `div` 4), 
                         az)
                  b_c = (cx, 
                        ((by + get_Y x + cy + (heights!!(mapping!!(((map_n_i `div` offs_x)*offs_x) + ((map_n_i + 1) `mod` offs_x))))) `div` 4),
                        (bz - cz) `div` 2 + cz)
                  d_c = ((cx - dx) `div` 2 + dx, 
                        ((cy + get_Y x + dy + (heights!!(mapping!!((map_n_i + offs_x) `mod` offs_y)))) `div` 4), 
                        dz)
                  a_d = (ax, 
                        ((ay + get_Y x + dy + (heights!!(mapping!!(((map_n_i `div` offs_x)*offs_x) + ((map_n_i - 1) `mod` offs_x))))) `div` 4),
                        (az - dz) `div` 2 + dz)
      
                  a = (ax, ay, az)
                  b = (bx, by, bz)
                  c = (cx, cy, cz)
                  d = (dx, dy, dz)
                  
                  -- Mapping between rectangle's position in the recursive list and the relative position in parent rectangle
                  f :: Int -> [Int]
                  f 0 = [0]
                  f n = (concat . (map (\y -> ((concat . map (\x -> [4*x, 4*x + 1])) y ++ 
                                               (concat . map (\x -> [4*x + 2, 4*x + 3])) y)))) 
                          (split_l (f (n - 1)) $ round $ sqrt $ fromIntegral $ length $ f (n - 1))
                            where split_l [] _ = []
                                  split_l xs n = take n xs : split_l (drop n xs) n

                  f_i n = map snd $ sort $ zip (f n) [0..]


getShapeList :: WorldObject -> [Shape]
getShapeList (Tree x) = x
getShapeList (Terrain x) = x
getShapeList (Object x) = x

translateObject :: WorldObject -> Point -> WorldObject
translateObject (Tree shapes) p = Tree (translate' p shapes)
translateObject (Object shapes) p = Object (translate' p shapes)

translate' = map . (flip translate)

------- GENERATED OBJECTS --------

tree1 :: WorldObject
tree1 = generateTree (Line (0, 0, 0) (0, 100, 0) bright_cyan) 60 4 (pi / 6.0) cyan
tree2 :: WorldObject
tree2 = generateTree (Line (0, 0, 0) (0, 120, 0) cyan) 50 4 (pi / 4.0) dark_cyan

lambda :: WorldObject
lambda = Object ((Polygon lambda_front white) : (Polygon lambda_back white) : lambda_side)

lambda_top_index = 3 :: Int

lambda_front :: [Point]
lambda_front = [(8,160,0),(24,192,0),(48,192,0),(56,176,0),(72,128,0),(88,64,0),(104,16,0),(112,16,0),(120,32,0),(136,32,0),(120,0,0),(96,0,0),(88,16,0),(72,64,0),(68,80,0),(32,0,0),(0,0,0),(40,80,0),(58,120,0),(56,128,0),(40,176,0),(32,176,0),(24,160,0)]
lambda_back :: [Point]
lambda_back = [(8,160,30),(24,192,30),(48,192,30),(56,176,30),(72,128,30),(88,64,30),(104,16,30),(112,16,30),(120,32,30),(136,32,30),(120,0,30),(96,0,30),(88,16,30),(72,64,30),(68,80,30),(32,0,30),(0,0,30),(40,80,30),(58,120,30),(56,128,30),(40,176,30),(32,176,30),(24,160,30)]
lambda_side = [Rectangle (8,160,0) (24,192,0) (24,192,30) (8,160,30) white, Rectangle (24,192,0) (48,192,0) (48,192,30) (24,192,30) white, Rectangle (48,192,0) (56,176,0) (56,176,30) (48,192,30) white, Rectangle (56,176,0) (72,128,0) (72,128,30) (56,176,30) white, Rectangle (72,128,0) (88,64,0) (88,64,30) (72,128,30) white, Rectangle (88,64,0) (104,16,0) (104,16,30) (88,64,30) white, Rectangle (104,16,0) (112,16,0) (112,16,30) (104,16,30) white, Rectangle (112,16,0) (120,32,0) (120,32,30) (112,16,30) white, Rectangle (120,32,0) (136,32,0) (136,32,30) (120,32,30) white, Rectangle (136,32,0) (120,0,0) (120,0,30) (136,32,30) white, Rectangle (120,0,0) (96,0,0) (96,0,30) (120,0,30) white, Rectangle (96,0,0) (88,16,0) (88,16,30) (96,0,30) white, Rectangle (88,16,0) (72,64,0) (72,64,30) (88,16,30) white, Rectangle (72,64,0) (68,80,0) (68,80,30) (72,64,30) white, Rectangle (68,80,0) (32,0,0) (32,0,30) (68,80,30) white, Rectangle (32,0,0) (0,0,0) (0,0,30) (32,0,30) white, Rectangle (0,0,0) (40,80,0) (40,80,30) (0,0,30) white, Rectangle (40,80,0) (58,120,0) (58,120,30) (40,80,30) white, Rectangle (58,120,0) (56,128,0) (56,128,30) (58,120,30) white, Rectangle (56,128,0) (40,176,0) (40,176,30) (56,128,30) white, Rectangle (40,176,0) (32,176,0) (32,176,30) (40,176,30) white, Rectangle (32,176,0) (24,160,0) (24,160,30) (32,176,30) white]

